package com.sun.xml.internal.ws.model;

import javax.xml.ws.Holder;
import com.sun.xml.internal.bind.v2.model.nav.Navigator;
import com.sun.xml.internal.bind.v2.model.annotation.AnnotationReader;
import com.sun.xml.internal.bind.v2.model.annotation.RuntimeInlineAnnotationReader;
import java.io.IOException;
import java.io.FileOutputStream;
import java.lang.reflect.TypeVariable;
import java.lang.reflect.WildcardType;
import java.lang.reflect.GenericArrayType;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.List;
import javax.xml.namespace.QName;
import java.lang.reflect.Method;
import com.sun.xml.internal.ws.org.objectweb.asm.MethodVisitor;
import com.sun.xml.internal.ws.org.objectweb.asm.FieldVisitor;
import java.util.Iterator;
import com.sun.xml.internal.ws.org.objectweb.asm.AnnotationVisitor;
import java.util.logging.Level;
import javax.xml.ws.WebServiceException;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlList;
import javax.xml.bind.annotation.XmlAttachmentRef;
import javax.xml.bind.annotation.adapters.XmlJavaTypeAdapter;
import javax.xml.bind.annotation.XmlMimeType;
import java.lang.annotation.Annotation;
import com.sun.xml.internal.ws.org.objectweb.asm.ClassWriter;
import java.util.Collection;
import java.util.logging.Logger;

public class WrapperBeanGenerator
{
    private static final Logger LOGGER;
    private static final WrapperBeanGenerator.FieldFactory FIELD_FACTORY;
    private static final AbstractWrapperBeanGenerator RUNTIME_GENERATOR;
    
    public WrapperBeanGenerator() {
    }
    
    private static byte[] createBeanImage(String className, String rootName, String rootNS, String typeName, String typeNS, Collection<WrapperBeanGenerator.Field> fields) throws Exception {
        ClassWriter cw = new ClassWriter(0);
        cw.visit(49, 33, replaceDotWithSlash(className), null, "java/lang/Object", null);
        AnnotationVisitor root = cw.visitAnnotation("Ljavax/xml/bind/annotation/XmlRootElement;", true);
        root.visit("name", rootName);
        root.visit("namespace", rootNS);
        root.visitEnd();
        AnnotationVisitor type = cw.visitAnnotation("Ljavax/xml/bind/annotation/XmlType;", true);
        type.visit("name", typeName);
        type.visit("namespace", typeNS);
        if (fields.size() > 1) {
            AnnotationVisitor propVisitor = type.visitArray("propOrder");
            for (WrapperBeanGenerator.Field field : fields) {
                propVisitor.visit("propOrder", field.fieldName);
            }
            propVisitor.visitEnd();
        }
        type.visitEnd();
        for (WrapperBeanGenerator.Field field2 : fields) {
            FieldVisitor fv = cw.visitField(1, field2.fieldName, field2.asmType.getDescriptor(), field2.getSignature(), null);
            for (Annotation ann : field2.jaxbAnnotations) {
                if ((ann instanceof XmlMimeType)) {
                    AnnotationVisitor mime = fv.visitAnnotation("Ljavax/xml/bind/annotation/XmlMimeType;", true);
                    mime.visit("value", ((XmlMimeType)ann).value());
                    mime.visitEnd();
                }
                else if ((ann instanceof XmlJavaTypeAdapter)) {
                    AnnotationVisitor ada = fv.visitAnnotation("Ljavax/xml/bind/annotation/adapters/XmlJavaTypeAdapter;", true);
                    ada.visit("value", getASMType(((XmlJavaTypeAdapter)ann).value()));
                    ada.visitEnd();
                }
                else if ((ann instanceof XmlAttachmentRef)) {
                    AnnotationVisitor att = fv.visitAnnotation("Ljavax/xml/bind/annotation/XmlAttachmentRef;", true);
                    att.visitEnd();
                }
                else if ((ann instanceof XmlList)) {
                    AnnotationVisitor list = fv.visitAnnotation("Ljavax/xml/bind/annotation/XmlList;", true);
                    list.visitEnd();
                }
                else if ((ann instanceof XmlElement)) {
                    AnnotationVisitor elem = fv.visitAnnotation("Ljavax/xml/bind/annotation/XmlElement;", true);
                    XmlElement xmlElem = (XmlElement)ann;
                    elem.visit("name", xmlElem.name());
                    elem.visit("namespace", xmlElem.namespace());
                    if (xmlElem.nillable()) {
                        elem.visit("nillable", Boolean.valueOf(true));
                    }
                    if (xmlElem.required()) {
                        elem.visit("required", Boolean.valueOf(true));
                    }
                    elem.visitEnd();
                }
                else {
                    throw new WebServiceException("Unknown JAXB annotation " + ann);
                }
            }
            fv.visitEnd();
        }
        MethodVisitor mv = cw.visitMethod(1, "<init>", "()V", null, null);
        mv.visitCode();
        mv.visitVarInsn(25, 0);
        mv.visitMethodInsn(183, "java/lang/Object", "<init>", "()V");
        mv.visitInsn(177);
        mv.visitMaxs(1, 1);
        mv.visitEnd();
        cw.visitEnd();
        if (WrapperBeanGenerator.LOGGER.isLoggable(Level.FINE)) {
            StringBuilder sb = new StringBuilder();
            sb.append("\n");
            sb.append("@XmlRootElement(name=").append(rootName).append(", namespace=").append(rootNS).append(")");
            sb.append("\n");
            sb.append("@XmlType(name=").append(typeName).append(", namespace=").append(typeNS);
            if (fields.size() > 1) {
                sb.append(", propOrder={");
                for (WrapperBeanGenerator.Field field3 : fields) {
                    sb.append(" ");
                    sb.append(field3.fieldName);
                }
                sb.append(" }");
            }
            sb.append(")");
            sb.append("\n");
            sb.append("public class ").append(className).append(" {");
            for (WrapperBeanGenerator.Field field3 : fields) {
                sb.append("\n");
                for (Annotation ann2 : field3.jaxbAnnotations) {
                    sb.append("\n    ");
                    if ((ann2 instanceof XmlMimeType)) {
                        sb.append("@XmlMimeType(value=").append(((XmlMimeType)ann2).value()).append(")");
                    }
                    else if ((ann2 instanceof XmlJavaTypeAdapter)) {
                        sb.append("@XmlJavaTypeAdapter(value=").append(getASMType(((XmlJavaTypeAdapter)ann2).value())).append(")");
                    }
                    else if ((ann2 instanceof XmlAttachmentRef)) {
                        sb.append("@XmlAttachmentRef");
                    }
                    else if ((ann2 instanceof XmlList)) {
                        sb.append("@XmlList");
                    }
                    else if ((ann2 instanceof XmlElement)) {
                        XmlElement xmlElem = (XmlElement)ann2;
                        sb.append("\n    ");
                        sb.append("@XmlElement(name=").append(xmlElem.name()).append(", namespace=").append(xmlElem.namespace());
                        if (xmlElem.nillable()) {
                            sb.append(", nillable=true");
                        }
                        if (xmlElem.required()) {
                            sb.append(", required=true");
                        }
                        sb.append(")");
                    }
                    else {
                        throw new WebServiceException("Unknown JAXB annotation " + ann2);
                    }
                }
                sb.append("\n    ");
                sb.append("public ");
                if (field3.getSignature() == null) {
                    sb.append(field3.asmType.getDescriptor());
                }
                else {
                    sb.append(field3.getSignature());
                }
                sb.append(" ");
                sb.append(field3.fieldName);
            }
            sb.append("\n\n}");
            WrapperBeanGenerator.LOGGER.fine(sb.toString());
        }
        return cw.toByteArray();
    }
    
    private static String replaceDotWithSlash(String name) {
        return name.replace('.', '/');
    }
    
    static Class createRequestWrapperBean(String className, Method method, QName reqElemName, ClassLoader cl) {
        if (WrapperBeanGenerator.LOGGER.isLoggable(Level.FINE)) {
            WrapperBeanGenerator.LOGGER.log(Level.FINE, "Request Wrapper Class : {0}", className);
        }
        List<WrapperBeanGenerator.Field> requestMembers = WrapperBeanGenerator.RUNTIME_GENERATOR.collectRequestBeanMembers(method);
        byte[] image;
        try {
            image = createBeanImage(className, reqElemName.getLocalPart(), reqElemName.getNamespaceURI(), reqElemName.getLocalPart(), reqElemName.getNamespaceURI(), requestMembers);
        }
        catch (Exception e) {
            throw new WebServiceException(e);
        }
        return Injector.inject(cl, className, image);
    }
    
    static Class createResponseWrapperBean(String className, Method method, QName resElemName, ClassLoader cl) {
        if (WrapperBeanGenerator.LOGGER.isLoggable(Level.FINE)) {
            WrapperBeanGenerator.LOGGER.log(Level.FINE, "Response Wrapper Class : {0}", className);
        }
        List<WrapperBeanGenerator.Field> responseMembers = WrapperBeanGenerator.RUNTIME_GENERATOR.collectResponseBeanMembers(method);
        byte[] image;
        try {
            image = createBeanImage(className, resElemName.getLocalPart(), resElemName.getNamespaceURI(), resElemName.getLocalPart(), resElemName.getNamespaceURI(), responseMembers);
        }
        catch (Exception e) {
            throw new WebServiceException(e);
        }
        return Injector.inject(cl, className, image);
    }
    
    private static com.sun.xml.internal.ws.org.objectweb.asm.Type getASMType(Type t) {
        assert t != null;
        if ((t instanceof Class)) {
            return com.sun.xml.internal.ws.org.objectweb.asm.Type.getType((Class)t);
        }
        if ((t instanceof ParameterizedType)) {
            ParameterizedType pt = (ParameterizedType)t;
            if ((pt.getRawType() instanceof Class)) {
                return com.sun.xml.internal.ws.org.objectweb.asm.Type.getType((Class)pt.getRawType());
            }
        }
        if ((t instanceof GenericArrayType)) {
            return com.sun.xml.internal.ws.org.objectweb.asm.Type.getType(FieldSignature.vms(t));
        }
        if ((t instanceof WildcardType)) {
            return com.sun.xml.internal.ws.org.objectweb.asm.Type.getType(FieldSignature.vms(t));
        }
        if ((t instanceof TypeVariable)) {
            TypeVariable<Object> tv = (TypeVariable)t;
            if ((tv.getBounds()[0] instanceof Class)) {
                return com.sun.xml.internal.ws.org.objectweb.asm.Type.getType((Class)tv.getBounds()[0]);
            }
        }
        throw new IllegalArgumentException("Not creating ASM Type for type = " + t);
    }
    
    static Class createExceptionBean(String className, Class exception, String typeNS, String elemName, String elemNS, ClassLoader cl) {
        return createExceptionBean(className, exception, typeNS, elemName, elemNS, cl, true);
    }
    
    static Class createExceptionBean(String className, Class exception, String typeNS, String elemName, String elemNS, ClassLoader cl, boolean decapitalizeExceptionBeanProperties) {
        Collection<WrapperBeanGenerator.Field> fields = WrapperBeanGenerator.RUNTIME_GENERATOR.collectExceptionBeanMembers(exception, decapitalizeExceptionBeanProperties);
        byte[] image;
        try {
            image = createBeanImage(className, elemName, elemNS, exception.getSimpleName(), typeNS, fields);
        }
        catch (Exception e) {
            throw new WebServiceException(e);
        }
        return Injector.inject(cl, className, image);
    }
    
    static void write(byte[] b, String className) {
        className = className.substring(className.lastIndexOf(".") + 1);
        try {
            FileOutputStream fo = new FileOutputStream(className + ".class");
            fo.write(b);
            fo.flush();
            fo.close();
        }
        catch (IOException e) {
            WrapperBeanGenerator.LOGGER.log(Level.INFO, "Error Writing class", e);
        }
    }
    
    static com.sun.xml.internal.ws.org.objectweb.asm.Type access$100(Type x0) {
        return getASMType(x0);
    }
    
    static {
        LOGGER = Logger.getLogger(WrapperBeanGenerator.class.getName());
        FIELD_FACTORY = new WrapperBeanGenerator.FieldFactory();
        RUNTIME_GENERATOR = new WrapperBeanGenerator.RuntimeWrapperBeanGenerator(new RuntimeInlineAnnotationReader(), Utils.REFLECTION_NAVIGATOR, WrapperBeanGenerator.FIELD_FACTORY);
    }
    
    private static final class RuntimeWrapperBeanGenerator extends AbstractWrapperBeanGenerator<Type, Class, Method, WrapperBeanGenerator.Field>
    {
        protected RuntimeWrapperBeanGenerator(AnnotationReader<Type, Class, ?, Method> annReader, Navigator<Type, Class, ?, Method> nav, AbstractWrapperBeanGenerator.BeanMemberFactory<Type, WrapperBeanGenerator.Field> beanMemberFactory) {
            super(annReader, nav, beanMemberFactory);
        }
        
        @Override
        protected Type getSafeType(Type type) {
            return type;
        }
        
        @Override
        protected Type getHolderValueType(Type paramType) {
            if ((paramType instanceof ParameterizedType)) {
                ParameterizedType p = (ParameterizedType)paramType;
                if (p.getRawType().equals(Holder.class)) {
                    return p.getActualTypeArguments()[0];
                }
            }
            return null;
        }
        
        @Override
        protected boolean isVoidType(Type type) {
            return type == Void.TYPE;
        }
    }
    
    private static final class FieldFactory implements AbstractWrapperBeanGenerator.BeanMemberFactory<Type, WrapperBeanGenerator.Field>
    {
        @Override
        public WrapperBeanGenerator.Field createWrapperBeanMember(Type paramType, String paramName, List<Annotation> jaxb) {
            return new WrapperBeanGenerator.Field(paramName, paramType, WrapperBeanGenerator.access$100(paramType), jaxb);
        }
    }
    
    private static class Field implements Comparable<WrapperBeanGenerator.Field>
    {
        private final Type reflectType;
        private final com.sun.xml.internal.ws.org.objectweb.asm.Type asmType;
        private final String fieldName;
        private final List<Annotation> jaxbAnnotations;
        
        Field(String paramName, Type paramType, com.sun.xml.internal.ws.org.objectweb.asm.Type asmType, List<Annotation> jaxbAnnotations) {
            this.reflectType = paramType;
            this.asmType = asmType;
            this.fieldName = paramName;
            this.jaxbAnnotations = jaxbAnnotations;
        }
        
        String getSignature() {
            if ((this.reflectType instanceof Class)) {
                return null;
            }
            if ((this.reflectType instanceof TypeVariable)) {
                return null;
            }
            return FieldSignature.vms(this.reflectType);
        }
        
        @Override
        public int compareTo(WrapperBeanGenerator.Field o) {
            return this.fieldName.compareTo(o.fieldName);
        }
    }
}
